@@ -8,8 +8,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution of
  * this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclWinChan.c,v 1.49.4.1 2008/05/23 21:10:45 andreas_kupries Exp $
  */
 
 #include "tclWinInt.h"
@@ -97,7 +95,7 @@ static void		FileThreadActionProc(ClientData instanceData,
 static int		FileTruncateProc(ClientData instanceData,
 			    Tcl_WideInt length);
 static DWORD		FileGetType(HANDLE handle);
-
+static int		NativeIsComPort(CONST TCHAR *nativeName);
 /*
  * This structure describes the channel type structure for file based IO.
  */
@@ -121,23 +119,6 @@ static Tcl_ChannelType fileChannelType = {
     FileThreadActionProc,	/* Thread action proc. */
     FileTruncateProc,		/* Truncate proc. */
 };
-
-#ifdef HAVE_NO_SEH
-/*
- * Unlike Borland and Microsoft, we don't register exception handlers by
- * pushing registration records onto the runtime stack. Instead, we register
- * them by creating an EXCEPTION_REGISTRATION within the activation record.
- */
-
-typedef struct EXCEPTION_REGISTRATION {
-    struct EXCEPTION_REGISTRATION* link;
-    EXCEPTION_DISPOSITION (*handler)(
-	    struct _EXCEPTION_RECORD*, void*, struct _CONTEXT*, void*);
-    void* ebp;
-    void* esp;
-    int status;
-} EXCEPTION_REGISTRATION;
-#endif
 
 /*
  *----------------------------------------------------------------------
@@ -491,7 +472,7 @@ FileSeekProc(
 
     oldPosHigh = 0;
     oldPos = SetFilePointer(infoPtr->handle, 0, &oldPosHigh, FILE_CURRENT);
-    if (oldPos == INVALID_SET_FILE_POINTER) {
+    if (oldPos == (LONG)INVALID_SET_FILE_POINTER) {
 	DWORD winError = GetLastError();
 
 	if (winError != NO_ERROR) {
@@ -503,7 +484,7 @@ FileSeekProc(
 
     newPosHigh = (offset < 0 ? -1 : 0);
     newPos = SetFilePointer(infoPtr->handle, offset, &newPosHigh, moveMethod);
-    if (newPos == INVALID_SET_FILE_POINTER) {
+    if (newPos == (LONG)INVALID_SET_FILE_POINTER) {
 	DWORD winError = GetLastError();
 
 	if (winError != NO_ERROR) {
@@ -566,7 +547,7 @@ FileWideSeekProc(
     newPosHigh = Tcl_WideAsLong(offset >> 32);
     newPos = SetFilePointer(infoPtr->handle, Tcl_WideAsLong(offset),
 	    &newPosHigh, moveMethod);
-    if (newPos == INVALID_SET_FILE_POINTER) {
+    if (newPos == (LONG)INVALID_SET_FILE_POINTER) {
 	DWORD winError = GetLastError();
 
 	if (winError != NO_ERROR) {
@@ -608,7 +589,7 @@ FileTruncateProc(
 
     oldPosHigh = 0;
     oldPos = SetFilePointer(infoPtr->handle, 0, &oldPosHigh, FILE_CURRENT);
-    if (oldPos == INVALID_SET_FILE_POINTER) {
+    if (oldPos == (LONG)INVALID_SET_FILE_POINTER) {
 	DWORD winError = GetLastError();
 	if (winError != NO_ERROR) {
 	    TclWinConvertError(winError);
@@ -623,7 +604,7 @@ FileTruncateProc(
     newPosHigh = Tcl_WideAsLong(length >> 32);
     newPos = SetFilePointer(infoPtr->handle, Tcl_WideAsLong(length),
 	    &newPosHigh, FILE_BEGIN);
-    if (newPos == INVALID_SET_FILE_POINTER) {
+    if (newPos == (LONG)INVALID_SET_FILE_POINTER) {
 	DWORD winError = GetLastError();
 	if (winError != NO_ERROR) {
 	    TclWinConvertError(winError);
@@ -904,6 +885,33 @@ TclpOpenFileChannel(
 	break;
     }
 
+    /*
+     * [2413550] Avoid double-open of serial ports on Windows
+     * Special handling for Windows serial ports by a "name-hint"
+     * to directly open it with the OVERLAPPED flag set.
+     */
+
+    if( NativeIsComPort(nativeName) ) {
+
+	handle = TclWinSerialOpen(INVALID_HANDLE_VALUE, nativeName, accessMode);
+	if (handle == INVALID_HANDLE_VALUE) {
+	    TclWinConvertError(GetLastError());
+	    if (interp != (Tcl_Interp *) NULL) {
+		Tcl_AppendResult(interp, "couldn't open serial \"",
+			TclGetString(pathPtr), "\": ",
+			Tcl_PosixError(interp), NULL);
+	    }
+	    return NULL;
+	}
+
+	/*
+	* For natively named Windows serial ports we are done.
+	*/
+	channel = TclWinOpenSerialChannel(handle, channelName,
+		channelPermissions);
+
+	return channel;
+    }
     /*
      * If the file is being created, get the file attributes from the
      * permissions argument, else use the existing file attributes.
@@ -954,11 +962,15 @@ TclpOpenFileChannel(
     switch (FileGetType(handle)) {
     case FILE_TYPE_SERIAL:
 	/*
+	 * Natively named serial ports "com1-9", "\\\\.\\comXX" are 
+	 * already done with the code above.
+	 * Here we handle all other serial port names.
+	 *
 	 * Reopen channel for OVERLAPPED operation. Normally this shouldn't
 	 * fail, because the channel exists.
 	 */
 
-	handle = TclWinSerialReopen(handle, nativeName, accessMode);
+	handle = TclWinSerialOpen(handle, nativeName, accessMode);
 	if (handle == INVALID_HANDLE_VALUE) {
 	    TclWinConvertError(GetLastError());
 	    if (interp != (Tcl_Interp *) NULL) {
@@ -1028,8 +1040,8 @@ Tcl_MakeFileChannel(
     int mode)			/* ORed combination of TCL_READABLE and
 				 * TCL_WRITABLE to indicate file mode. */
 {
-#ifdef HAVE_NO_SEH
-    EXCEPTION_REGISTRATION registration;
+#if defined(HAVE_NO_SEH) && !defined(_WIN64)
+    TCLEXCEPTION_REGISTRATION registration;
 #endif
     char channelName[16 + TCL_INTEGER_SPACE];
     Tcl_Channel channel = NULL;
@@ -1094,12 +1106,7 @@ Tcl_MakeFileChannel(
 	 */
 
 	result = 0;
-#ifndef HAVE_NO_SEH
-	__try {
-	    CloseHandle(dupedHandle);
-	    result = 1;
-	} __except (EXCEPTION_EXECUTE_HANDLER) {}
-#else
+#if defined(HAVE_NO_SEH) && !defined(_WIN64)
 	/*
 	 * Don't have SEH available, do things the hard way. Note that this
 	 * needs to be one block of asm, to avoid stack imbalance; also, it is
@@ -1115,7 +1122,7 @@ Tcl_MakeFileChannel(
 	    "movl       %[dupedHandle], %%ebx"          "\n\t"
 
 	    /*
-	     * Construct an EXCEPTION_REGISTRATION to protect the call to
+	     * Construct an TCLEXCEPTION_REGISTRATION to protect the call to
 	     * CloseHandle.
 	     */
 
@@ -1129,7 +1136,7 @@ Tcl_MakeFileChannel(
 	    "movl       $0,             0x10(%%edx)"    "\n\t" /* status */
 
 	    /*
-	     * Link the EXCEPTION_REGISTRATION on the chain.
+	     * Link the TCLEXCEPTION_REGISTRATION on the chain.
 	     */
 
 	    "movl       %%edx,          %%fs:0"         "\n\t"
@@ -1142,7 +1149,7 @@ Tcl_MakeFileChannel(
 	    "call       _CloseHandle@4"                 "\n\t"
 
 	    /*
-	     * Come here on normal exit. Recover the EXCEPTION_REGISTRATION
+	     * Come here on normal exit. Recover the TCLEXCEPTION_REGISTRATION
 	     * and put a TRUE status return into it.
 	     */
 
@@ -1152,7 +1159,7 @@ Tcl_MakeFileChannel(
 	    "jmp        2f"                             "\n"
 
 	    /*
-	     * Come here on an exception. Recover the EXCEPTION_REGISTRATION
+	     * Come here on an exception. Recover the TCLEXCEPTION_REGISTRATION
 	     */
 
 	    "1:"                                        "\t"
@@ -1161,7 +1168,7 @@ Tcl_MakeFileChannel(
 
 	    /*
 	     * Come here however we exited. Restore context from the
-	     * EXCEPTION_REGISTRATION in case the stack is unbalanced.
+	     * TCLEXCEPTION_REGISTRATION in case the stack is unbalanced.
 	     */
 
 	    "2:"                                        "\t"
@@ -1179,7 +1186,15 @@ Tcl_MakeFileChannel(
 	    "%eax", "%ebx", "%ecx", "%edx", "%esi", "%edi", "memory"
 	    );
 	result = registration.status;
-
+#else
+#ifndef HAVE_NO_SEH
+	__try {
+#endif
+	    CloseHandle(dupedHandle);
+	    result = 1;
+#ifndef HAVE_NO_SEH
+	} __except (EXCEPTION_EXECUTE_HANDLER) {}
+#endif
 #endif
 	if (result == FALSE) {
 	    return NULL;
@@ -1223,7 +1238,7 @@ TclpGetDefaultStdChannel(
     HANDLE handle;
     int mode = -1;
     char *bufMode = NULL;
-    DWORD handleId = (DWORD)INVALID_HANDLE_VALUE;
+    DWORD handleId = (DWORD)-1;
 				/* Standard handle to retrieve. */
 
     switch (type) {
@@ -1336,7 +1351,7 @@ TclWinOpenFileChannel(
     infoPtr->flags = appendMode;
     infoPtr->handle = handle;
     infoPtr->dirty = 0;
-    wsprintfA(channelName, "file%lx", (int) infoPtr);
+    sprintf(channelName, "file%" TCL_I_MODIFIER "x", (size_t)infoPtr);
 
     infoPtr->channel = Tcl_CreateChannel(&fileChannelType, channelName,
 	    (ClientData) infoPtr, permissions);
@@ -1491,6 +1506,122 @@ FileGetType(
 
     return type;
 }
+
+ /*
+ *----------------------------------------------------------------------
+ *
+ * NativeIsComPort --
+ *
+ *	Determines if a path refers to a Windows serial port.
+ *	A simple and efficient solution is to use a "name hint" to detect 
+ *      COM ports by their filename instead of resorting to a syscall 
+ *	to detect serialness after the fact.
+ *	The following patterns cover common serial port names:
+ *	    COM[1-9]:?
+ *	    //./COM[0-9]+
+ *	    \\.\COM[0-9]+
+ *
+ * Results:
+ *	1 = serial port, 0 = not.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static int
+NativeIsComPort(
+    const TCHAR *nativePath)	/* Path of file to access, native encoding. */
+{
+    /*
+     * Use wide-char or plain character case-insensitive comparison
+     */
+    if (tclWinProcs->useWide) {
+	const WCHAR *p = (const WCHAR *) nativePath;
+	int i, len = wcslen(p);
+
+	/*
+	 * 1. Look for com[1-9]:?
+	 */
+
+	if ( (len >= 4) && (len <= 5) 
+		&& (_wcsnicmp(p, L"com", 3) == 0) ) {
+	    /*
+	    * The 4th character must be a digit 1..9 optionally followed by a ":"
+	    */
+
+	    if ( (p[3] < L'1') || (p[3] > L'9') ) {
+		return 0;
+	    }
+	    if ( (len == 5) && (p[4] != L':') ) {
+		return 0;
+	    }
+	    return 1;
+	}
+
+	/*
+	 * 2. Look for //./com[0-9]+ or \\.\com[0-9]+
+	 */
+
+	if ( (len >= 8) && ( 
+		   (_wcsnicmp(p, L"//./com", 7) == 0)
+		|| (_wcsnicmp(p, L"\\\\.\\com", 7) == 0) ) )
+	{
+	    /*
+	    * Charaters 8..end must be a digits 0..9
+	    */
+
+	    for ( i=7; i<len; i++ ) {
+		if ( (p[i] < '0') || (p[i] > '9') ) {
+		    return 0;
+		}
+	    }
+	    return 1;
+	}
+
+    } else {
+	const char *p = (const char *) nativePath;
+	int   i, len = strlen(p);
+
+	/*
+	 * 1. Look for com[1-9]:?
+	 */
+
+	if ( (len >= 4) && (len <= 5) 
+		&& (strnicmp(p, "com", 3) == 0) ) {
+	    /*
+	    * The 4th character must be a digit 1..9 optionally followed by a ":"
+	    */
+
+	    if ( (p[3] < '1') || (p[3] > '9') ) {
+		return 0;
+	    }
+	    if ( (len == 5) && (p[4] != ':') ) {
+		return 0;
+	    }
+	    return 1;
+	}
+
+	/*
+	 * 2. Look for //./com[0-9]+ or \\.\com[0-9]+
+	 */
+
+	if ( (len >= 8) && ( 
+		   (strnicmp(p, "//./com", 7) == 0)
+		|| (strnicmp(p, "\\\\.\\com", 7) == 0) ) )
+	{
+	    /*
+	    * Charaters 8..end must be a digits 0..9
+	    */
+
+	    for ( i=7; i<len; i++ ) {
+		if ( (p[i] < '0') || (p[i] > '9') ) {
+		    return 0;
+		}
+	    }
+	    return 1;
+	}
+    }
+    return 0;
+}
 
 /*
  * Local Variables:
